﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Linq.Expressions;
using System.Windows;

namespace DragDropFrameworkPOC.BaseClasses
{
  /// <summary>
  /// A base class for all view models.
  /// <remarks>
  /// In order to implement the Dispose pattern by any implemented class. Plese refer the below code snippet.
  /// <para/>
  /// The following pattern should be followed by a view model class which derives (and should derive from) ViewModelBase class.
  /// <example>
  ///     <code>
  ///         public class $DerivedClassName$: ViewModelBase
  ///          {
  ///              private bool _disposed = false;
  ///              protected override void Dispose(bool disposing)
  ///              {
  ///                  if (!_disposed)
  ///                  {
  ///                      if (disposing)
  ///                      {
  ///                          // Release managed resources.
  ///                      }
  ///                      // Release unmanaged resources.
  ///                      // Set large fields to null.
  ///                      // Call Dispose on your base class.
  ///                      _disposed = true;
  ///                  }
  ///                  base.Dispose(disposing);
  ///              }
  ///              // The derived class does not have a Finalize method
  ///              // or a Dispose method without parameters because it inherits
  ///              // them from the base class.
  ///          }
  ///     </code>
  /// </example>
  /// </remarks>
  /// </summary>
  public abstract class ViewModelBase : INotifyPropertyChanged, IDataErrorInfo, IDisposable, IWeakEventListener
  {
    #region Members Variables

    private PropertyChangedEventHandler _propertyChanged;
    private string _displayName;
    private bool _disposed = false;
    private string _error;

    #endregion

    #region INotifyPropertyChanged Members

    /// <summary>
    /// Gets or sets the property changed event.
    /// </summary>
    public event PropertyChangedEventHandler PropertyChanged
    {
      add { _propertyChanged += value; }
      remove { _propertyChanged -= value; }
    }

    #endregion

    #region Constructor

    /// <summary>
    /// Initializes a new instance of <see cref="ViewModelBase"/>.
    /// </summary>
    protected ViewModelBase()
    {
      PropertyNames = new List<string>();
    }

    #endregion

    #region IDisposable Members

    /// <summary>
    /// Cleans up the memory.
    /// </summary>
    public void Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }
    #endregion

    #region IDataErrorInfo Members

    /// <summary>
    /// Return error of the ViewModel.
    /// </summary>
    public string Error
    {
      get { return _error; }
      set
      {
        if (value != _error)
        {
          _error = value;
          OnPropertyChange(() => Error);
        }
      }

    }

    /// <summary>
    /// An Indexed based error.
    /// </summary>
    /// <param name="columnName">property name.</param>
    /// <returns></returns>
    public string this[string columnName]
    {
      get { return VerifyPropertyValue(columnName); }
    }

    #endregion

    #region Public Members

    /// <summary>
    /// Gets if the ViewModel is in valid state.
    /// </summary>
    public bool IsValid
    {
      get { return IsValidMethod(); }
    }

    /// <summary>
    /// Gets or sets the DiplayName property of the ViewModel instance.
    /// A handy proprty for workspace headings or default string representation.
    /// </summary>
    public string DisplayName
    {
      get { return _displayName; }
      set
      {
        if (_displayName != value)
        {
          _displayName = value;
          OnPropertyChange(() => DisplayName);
        }
      }
    }

    /// <summary>
    /// Call the method to initialize the instance of <see cref="ViewModelBase"/>.
    /// <remarks>Internally this just calls the overridable method <see cref="OnInitialize"/> which would be overriden by any subclass.</remarks>
    /// </summary>
    public void Initialize()
    {
      OnInitialize();
    }

    #endregion

    #region Protected/Overridable Members

    /// <summary>
    /// Override the method for any memory cleaning.
    /// </summary>
    /// <param name="disposing"></param>
    protected virtual void Dispose(bool disposing)
    {
      if (!_disposed)
      {
        if (disposing)
        {
          // Free other state (managed objects).
        }
        // Free your own state (unmanaged objects).
        // Set large fields to null.
        _disposed = true;
      }
    }

    /// <summary>
    /// Fires the propertyChanged event for propertyName.
    /// </summary>
    /// <typeparam name="TProp"></typeparam>
    /// <param name="expression"></param>
    protected virtual void OnPropertyChange<TProp>(Expression<Func<TProp>> expression)
    {
      var propertyName = GetPropertyName(expression);
      var propertyChanged = _propertyChanged;
      if (null != propertyChanged)
      {
        propertyChanged(this, new PropertyChangedEventArgs(propertyName));
       // RefreshAllCommands();
      }
    }

    /// <summary>
    /// The property names that need to be verified.
    /// </summary>
    protected IList<string> PropertyNames;

    /// <summary>
    /// Verifies if a property values is valid.
    /// </summary>
    /// <param name="propertyName">property name to be validated.</param>
    /// <returns>error string.</returns>
    protected abstract string VerifyPropertyValue(string propertyName);

    /// <summary>
    /// Refreshes all the commands that exist in the <see cref="CommandCollection"/>.
    /// This evaluates all the commands for CanExecute state.
    /// </summary>
    //protected void RefreshAllCommands()
    //{
    //  CommandCollection.ForEach(cmd => cmd.RaiseCanExecuteChanged());
    //}

    /// <summary>
    /// A utility mehod for thread safe method invoking.
    /// </summary>
    /// <param name="action">The method that needs to be called in a thread safe manner..</param>
    protected void ThreadSafeInvoke(Action action)
    {
      //if (Utility.CurrentDispatcher.CheckAccess())
      //{
      //  action();
      //}
      //else
      //{
      //  Utility.CurrentDispatcher.Invoke(action, null);
      //}
    }

    /// <summary>
    /// Gets the property name for property.
    /// </summary>
    /// <typeparam name="TProp"></typeparam>
    /// <param name="expression">lambda expression pointing to property.</param>
    /// <returns>propertyName.</returns>
    protected string GetPropertyName<TProp>(Expression<Func<TProp>> expression)
    {
      var body = expression.Body as MemberExpression;
      if (body == null)
        throw new ArgumentException("'expression' should be a member expression");
      return body.Member.Name;
    }

    /// <summary>
    /// An overridable method to place some custom initialization logic.
    /// This method is called through <see cref="Initialize"/> method.
    /// </summary>
    protected virtual void OnInitialize()
    {
      //Place logic for initialization.
    }

    /// <summary>
    /// Derived class implementation to be done for implementing WeakEventPattern.
    /// </summary>
    /// <param name="managerType"></param>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    /// <returns></returns>
    protected virtual bool OnReceiveWeakEvent(Type managerType, object sender, EventArgs e)
    {
      throw new NotImplementedException();
    }

    #endregion

    #region Private Members

    /// <summary>
    /// Return false if any property of the view model returns a null error.
    /// </summary>
    /// <returns></returns>
    protected virtual bool IsValidMethod()
    {
      if (null != PropertyNames)
      {
        return !PropertyNames.Any(p => VerifyPropertyValue(p) != null);
      }
      return true;
    }

    #endregion

    #region Finalizer

    /// <summary>
    /// The finalizer.
    /// </summary>
    ~ViewModelBase()
    {
      Dispose(false);
    }

    #endregion

    #region IWeakEventListener Implementation

    /// <summary>
    /// Used by a WeakEventManager to notify of an event.
    /// </summary>
    /// <param name="managerType"></param>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    /// <returns></returns>
    public bool ReceiveWeakEvent(Type managerType, object sender, EventArgs e)
    {
      return OnReceiveWeakEvent(managerType, sender, e);
    }

    #endregion

  }
}
